r"""
Utilities based on the Taylor expansion of the functions of the form
:math:`\frac{(1 + z)^k}{E(z)}` for flat wCDM cosmology.
These functions and their integral are estimated using the Pade approximation.
"""

import numpy as xp
from scipy.linalg import toeplitz

from .utils import autodoc

__all__ = [
    "analytic_integral",
    "flat_wcdm_pade_coefficients",
    "flat_wcdm_taylor_expansion",
]


def pade(an, m, n=None):
    """
    Return Pade approximation to a polynomial as the ratio of two polynomials.

    Parameters
    ----------
    an : (N,) array_like
        Taylor series coefficients.
    m : int
        The order of the returned approximating polynomial `q`.
    n : int, optional
        The order of the returned approximating polynomial `p`. By default,
        the order is ``len(an)-1-m``.

    Returns
    -------
    p, q : Polynomial class
        The Pade approximation of the polynomial defined by `an` is
        ``p(x)/q(x)``.

    Notes
    -----
    This code has been slightly edited from the scipy implementation to:

    - Use xp instead of np to support multiple backends
    - Directly use the fact that part of the matrix is Toeplitz

    """
    an = xp.asarray(an)
    if n is None:
        n = len(an) - 1 - m
        if n < 0:
            raise ValueError("Order of q <m> must be smaller than len(an)-1.")
    if n < 0:
        raise ValueError("Order of p <n> must be greater than 0.")
    N = m + n
    if N > len(an) - 1:
        raise ValueError("Order of q+p <m+n> must be smaller than len(an).")
    an = an[: N + 1]
    Akj = xp.eye(N + 1, n + 1, dtype=an.dtype)
    Bkj = toeplitz(xp.r_[0.0, -an[:-1]], xp.zeros(m))
    Ckj = xp.hstack((Akj, Bkj))
    pq = xp.linalg.solve(Ckj, an)
    p = pq[: n + 1]
    q = xp.r_[1.0, pq[n + 1 :]]
    return p[::-1], q[::-1]


@autodoc
def flat_wcdm_taylor_expansion(w0, zpower=0):
    r"""
    Taylor coefficients expansion of :math:`E(z)` as as a function
    of :math:`w_0` and an arbitrary power :math:`k` of :math:`1 + z`.

    .. math::

        F(x; w_0; k) = 2\sum_{{n=0}}^{{\infty}} \binom{{-\frac{{1}}{{2}}}}{{n}}
        \frac{{x^n}}{{\left(1 - 2k - 6nw_0\right)}}

    We include terms up to :math:`n=16`.

    Parameters
    ----------
    {w0}
    {zpower}

    Returns
    -------
    array_like
        The Taylor expansion coefficients.
    """
    power = 1 - 2 * zpower
    return xp.array(
        [
            w0**0 / power,
            -1 / (2 * (power - 6 * w0)),
            3 / (8 * (power - 12 * w0)),
            -5 / (16 * (power - 18 * w0)),
            35 / (128 * (power - 24 * w0)),
            -63 / (256 * (power - 30 * w0)),
            231 / (1024 * (power - 36 * w0)),
            -429 / (2048 * (power - 42 * w0)),
            6435 / (32768 * (power - 48 * w0)),
            -12155 / (65536 * (power - 54 * w0)),
            46189 / (262144 * (power - 60 * w0)),
            -88179 / (524288 * (power - 66 * w0)),
            676039 / (4194304 * (power - 72 * w0)),
            -1300075 / (8388608 * (power - 78 * w0)),
            5014575 / (33554432 * (power - 84 * w0)),
            -9694845 / (67108864 * (power - 90 * w0)),
            300540195 / (268435456 * (power - 96 * w0)),
        ]
    )


@autodoc
def flat_wcdm_pade_coefficients(w0=-1, zpower=0):
    """
    Compute the Pade coefficients as described in arXiv:1111.6396.
    We make two primary changes:

    - allow a variable dark energy equation of state :math:`w_0` by changing
      the definition of :math:`x`.
    - include more terms (17) in the Taylor expansion.

    Parameters
    ----------
    {w0}

    Returns
    -------
    p, q: array_like
        The Pade coefficients
    """
    coeffs = flat_wcdm_taylor_expansion(w0, zpower=zpower)
    p, q = pade(coeffs, len(coeffs) // 2, len(coeffs) // 2)
    return p, q


@autodoc
def indefinite_integral(z, Om0, w0=-1, zpower=0):
    r"""
    Compute the Pade approximation to :math:`(1+z)^k / E(z)` as described in
    arXiv:1111.6396. We extend it to include a variable dark energy
    equation of state, other integrands via powers of :math:`1 + z` and
    include more terms in the expansion

    .. math::

        I(z; \Omega_{{m, 0}}, w_0) = \frac{{-2 (1 + z)^{{\frac{{1}}{{2}}-k}}}}
        {{\Omega_{{m, 0}}^{{\frac{{1}}{{2}}}}}} \Phi(z; \Omega_{{m, 0}}, w_0).

    .. math::

        \Phi(z; \Omega_{{m, 0}}, w_0) =
        \frac{{\sum_i^n \alpha_i x^i}}{{1 + \sum_{{j=1}}^m \beta_j x^j}}.

    Here the expansion is in terms of

    .. math::

        x = \left(\frac{{1 - \Omega_{{m, 0}}}}{{\Omega_{{m, 0}}}}\right)
        (1 + z)^{{3 w_0}}.

    In practice we use :math:`m=n=7` whereas Adachi and Kasai use :math:`m=n=3`.

    Parameters
    ----------
    {z}
    {Om0}
    {w0}

    Returns
    -------
    I: array_like
        The indefinite integral of :math:`(1+z)^k / E(z)`
    """
    x = (1 - Om0) / Om0 * (1 + z) ** (3 * w0)
    p, q = flat_wcdm_pade_coefficients(w0=w0, zpower=zpower)
    normalization = -2 / Om0**0.5 / (1 + z) ** (0.5 - zpower)
    return normalization * xp.polyval(p, x) / xp.polyval(q, x)


@autodoc
def analytic_integral(z, Om0, w0=-1, zpower=0):
    r"""
    Compute an integral of the form
    :math:`\int_{{\infty}}^z \frac{{(1+z)^k}}{{E(z)}}` using

    .. math::

        f(z; \Omega_{{m, 0}}, w_0) = \int_{{\infty}}^{{z}}
        \frac{{dz' (1 + z')^k}}{{E(z'; \Omega_{{m, 0}}, w_0)}}
        = \frac{{-2\Phi(z; \Omega_{{m, 0}}, w_0) (1+z)^k }}
        {{\sqrt{{\Omega_{{m, 0}}(1 + z)}}}}.

    The integral is approximated using the Pade approximation and is up
    to a factor the term in the braces in (1.1) of Adachi and Kasai.

    Parameters
    ----------
    {z}
    {Om0}
    {w0}

    Returns
    -------
    integral: array_like
        The integral of :math:`1 / E(z)` from :math:`\infty` to :math:`z`
    """
    kwargs = dict(Om0=Om0, w0=w0, zpower=zpower)
    return indefinite_integral(z, **kwargs) - indefinite_integral(0, **kwargs)



import numpy as np
import pytest
from gwpopulation.utils import to_numpy
from scipy.interpolate import pade as spade

from ..taylor import pade as wpade


def test_pade(backend):
    taylor = np.random.uniform(0, 1, 10)
    scoeffs = spade(taylor, 4, 3)
    wcoeffs = wpade(taylor, 4, 3)
    assert max(abs(np.array(scoeffs[0]) - to_numpy(wcoeffs[0]))) < 1e-13
    assert max(abs(np.array(scoeffs[1]) - to_numpy(wcoeffs[1]))) < 1e-13


def test_pade_raises_n_less_0():
    taylor = np.random.uniform(0, 1, 10)
    with pytest.raises(ValueError):
        wpade(taylor, 4, -3)


def test_pade_n_is_none():
    taylor = np.random.uniform(0, 1, 10)
    scoeffs = spade(taylor, 4)
    wcoeffs = wpade(taylor, 4)
    assert max(abs(np.array(scoeffs[0]) - to_numpy(wcoeffs[0]))) < 1e-13
    assert max(abs(np.array(scoeffs[1]) - to_numpy(wcoeffs[1]))) < 1e-13
    with pytest.raises(ValueError):
        wpade(taylor, 20)


def test_pade_raises_m_n_too_large():
    taylor = np.random.uniform(0, 1, 10)
    with pytest.raises(ValueError):
        wpade(taylor, 20, 20)



"""
Helper functions that are not directly relevant to cosmology.
"""

import numpy as xp

_cosmology_docstrings_ = dict(
    z="""z: array_like
        Redshift""",
    Om0="""Om0: array_like
        The matter density fraction""",
    w0="""w0: array_like
        The (constant) equation of state parameter for dark energy""",
    H0="""H0: array_like
        The Hubble constant in km/s/Mpc""",
    zmin="""zmin: float
        The minimum redshift used in the conversion from distance to redshift,
        default=1e-4""",
    zmax="""zmax: float
        The maximum redshift used in the conversion from distance to redshift,
        default=100""",
    name="""name: str
        The name for the cosmology, mostly used for fixed instances""",
    meta="""meta: dict
        Additional metadata describing the cosmology, e.g., citation
        information""",
    m1="""m1: array_like
        The primary mass in the source frame""",
    m2="""m2: array_like
        The secondary mass in the source frame""",
    m1z="""m1z: array_like
        The primary mass in the detector frame""",
    m2z="""m2z: array_like
        The secondary mass in the detector frame""",
    dL="""dL: array_like
        The luminosity distance in Mpc""",
    zpower="""zpower: array_like
        The power of the redshift dependence of the distance integrand
        (:math:`k`)""",
)

GYR_KM_PER_S_MPC = 977.7922216807891  # for converting km/s/Mpc to Gyr
SPEED_OF_LIGHT_KM_PER_S = 299792.4580  # km/s


def autodoc(func):
    """
    Simple decorator to mark that a docstring needs formatting
    """
    func.__doc__ = func.__doc__.format(**_cosmology_docstrings_)
    return func


def method_autodoc(alt=None):
    """
    Simple decorator to mark that a docstring needs formatting.
    This will strip the class level attributes of :code:`FlatwCDM`
    from the dosctring and allow a docstring to be taken from
    another function.
    """

    def new_wrapper(func):
        def _strip_wcdm_parameters(doc):
            for key in ["H0", "Om0", "w0"]:
                doc = doc.replace(_cosmology_docstrings_[key], "")
            return doc

        if alt is not None:
            doc = alt.__doc__
        else:
            doc = func.__doc__
        doc = _strip_wcdm_parameters(doc)
        func.__doc__ = doc
        return func

    return new_wrapper


def maybe_jit(func, *args, **kwargs):
    """
    A decorator to jit the function if using jax.

    This also allows arbitrary arguments to be passed through,
    e.g., to specify static arguments.

    This function is pretty useful and so might make it into
    :code:`gwpopulation` regardless of cosmology.
    """
    if "jax" in xp.__name__:
        from jax import jit

        return jit(func, *args, **kwargs)
    return func



import numpy as np
import pytest
from astropy import cosmology
from gwpopulation.utils import to_numpy

from .. import wcosmo

funcs = [
    "luminosity_distance",
    "comoving_volume",
    "comoving_distance",
    "efunc",
    "inv_efunc",
    "differential_comoving_volume",
    "absorption_distance",
    "lookback_time",
]


@pytest.mark.parametrize("func", funcs)
def test_redshift_function(cosmo, func, backend):
    ours = wcosmo.available[cosmo]
    theirs = getattr(cosmology, cosmo)

    redshifts = np.linspace(1e-3, 10, 1000)

    ours = to_numpy(getattr(ours, func)(redshifts))
    theirs = getattr(theirs, func)(redshifts)
    if hasattr(theirs, "value"):
        theirs = theirs.value
    assert max(abs(ours - theirs) / theirs) < 1e-2


@pytest.mark.parametrize("func", funcs[:3])
def test_z_at_value(cosmo, func, backend):
    from gwpopulation.utils import xp

    ours = getattr(wcosmo.available[cosmo], func)
    theirs = getattr(getattr(cosmology, cosmo), func)

    redshifts = np.linspace(1e-3, 10, 10)
    vals = theirs(redshifts)

    if hasattr(vals, "value"):
        ovals = xp.asarray(vals.value)
    else:
        ovals = xp.asarray(vals)

    ours = wcosmo.z_at_value(ours, ovals)
    theirs = cosmology.z_at_value(theirs, vals).value
    assert max(abs(ours - theirs) / theirs) < 1e-2



import gwpopulation
import pytest

from ..wcosmo import FlatLambdaCDM, available


@pytest.fixture(params=["numpy", "jax", "cupy"])
def backend(request):
    pytest.importorskip(request.param)
    gwpopulation.set_backend(request.param)
    return request.param


@pytest.fixture
def npy():
    return gwpopulation.set_backend("numpy")


@pytest.fixture(params=available.keys())
def cosmo(request):
    ours = available[request.param]
    if not isinstance(ours, FlatLambdaCDM):
        pytest.skip()
    return request.param



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import importlib.metadata

project = "wcosmo"
copyright = "2024, Colm Talbot, Amanda Farah"
author = "Colm Talbot, Amanda Farah"
release = importlib.metadata.version("wcosmo").split("+")[0]

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
    "sphinx.ext.autosummary",
    "numpydoc",
    "nbsphinx",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]
exclude_patterns = []

root_doc = "index"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"

autosummary_generate = True
numpydoc_show_class_members = False
